@@ -23,9 +23,9 @@ use POE::Component::IKC::Channel;
 require Exporter;
 @ISA = qw(Exporter);
 @EXPORT = qw(create_ikc_client);
-$VERSION = '0.02';
+$VERSION = '0.03';
 
-sub DEBUG { 1 }
+sub DEBUG { 0 }
 
 
 ###############################################################################
@@ -109,7 +109,8 @@ POE::Component::IKC::Client - POE Inter-Kernel Communication server
         ip=>$ip, 
         port=>$port,
         name=>"Client$$",
-        on_connect=>\&create_sessions);
+        on_connect=>\&create_sessions,
+        subscribe=>[qw(poe:/*/timserver)],);
     ...
     $poe_kernel->run();
 
@@ -150,14 +151,16 @@ Code ref that is called when the connection has been made to the foreign
 kernel.  Normaly, you would use this to start the sessions that post events
 to foreign kernels.
 
+=item C<subscribe>
+
+Array ref of specifiers (either foreign sessions, or foreign states) that
+you want to subscribe to.  on_connect will only be called when IKC has
+managed to subscribe to each specifier.  If it can't, it will die().
+
 =back
 
 =head1 BUGS
 
-Currently, a foreign POE kernel will have access to all events in sessions
-that have aliases within the local kernel.  This is more then enough
-rope to shot yourself in the foot. :)
-
 =head1 AUTHOR
 
 Philip Gwyn, <fil@pied.nu>
